#! /usr/bin/env python
 
# ROS base controller node  
# MM control over BLE Nordic nRF5* UART
# Wes Freeman 2018. 
# Uses Adafruit library by Tony DiCola
# adapted from code at https://admin.insights.ubuntu.com/2018/03/09/your-first-robot-the-driver-4-5/
# uses lib https://github.com/cmcqueen/cobs-python
# Python 2.7
 
import rospy
from geometry_msgs.msg import Twist
 
import numpy as np
import binascii
import time
from struct import * # for 'pack'
from cobs import cobs
import Adafruit_BluefruitLE
from Adafruit_BluefruitLE.services import UART
from pick import pick # for UART list
 

class globs():
    velX = 0
    velY = 0
    velTheta = 0
 
 
 
def calc_checksum(cmd):
 
    ints = [ord(char) for char in cmd]
    checksum = sum(ints)  
    checksum %= 256 # roll over if bigger than 8b max
    checksum ^= 256 # twos complement
    checksum += 1 # why add 1?
    checksum %= 256 # roll over if bigger than 8b max
    #print ('checksum:',checksum)
    return checksum
 
 
 
 
# MM_command produces hex for MM command from signed int inputs
# this uses actual byte stuffing method
def MM_command(intSeq):
 
    #print('intSeq:', intSeq)
    # build the command, append vars in hex string '\xff' format
    cmd = ''
    for x in intSeq:
        cmd += pack('b',x) # b means treat as signed +- 128
 
    #print ('cmd string:', cmd)
 
    # COBS before the checksum and terminal byte
    cmd = cobs.encode(cmd)
    #print ('cobs encode: ', cmd)
 
    # checksum
    checksum = calc_checksum(cmd)
    cmd += chr(checksum) # chr hexifies 0-255
 
    # terminator
    cmd += chr(0)
 
    # convert to hex literal string without \x
    cmd = binascii.hexlify(cmd)
 
    return cmd
 
 
 
 
 
def main():
 
    # Clear any cached data because both bluez and CoreBluetooth have issues with
    # caching data and it going stale.
    ble.clear_cached_data()
 
    # Get the first available BLE network adapter and make sure it's powered on.
    adapter = ble.get_default_adapter()
    adapter.power_on()
    print('Using adapter: {0}'.format(adapter.name))
 
    # Disconnect any currently connected UART devices.  Good for cleaning up and
    # starting from a fresh state.
    print('Disconnecting any connected UART devices...')
    UART.disconnect_devices()
 
    # Scan for UART devices.
    print('Searching for UART devices...')
    adapter.start_scan()
 
    known_uarts = set() 
 
    time_end = time.time() + 10 # search for x secs
    while time.time() < time_end:
        # Call UART.find_devices to get a list of any UART devices that
        # have been found.  This call will quickly return results and does
        # not wait for devices to appear.
        found = set(UART.find_devices())
        # Check for new devices that haven't been seen yet and print out
        # their name and ID (MAC address on Linux, GUID on OSX).
        new = found - known_uarts
        for device in new:
            print('Found UART: {0} [{1}]'.format(device.name, device.id))
        known_uarts.update(new)
        # Sleep for a second and see if new devices have appeared.
        time.sleep(1.0)
 
    
    # Make sure scanning is stopped before exiting.
    adapter.stop_scan()
    devNames = []
    UARTlist = []
    for device in known_uarts:
        devNames.append(device.name)
        UARTlist.append(device) # copy set to list to make it inexable
 
    title = 'Please choose Mekamon to control: '
    option, index = pick(devNames, title)
    print option
    print index
 
    device = UARTlist[index]
 
    print('Connecting to device...')
    device.connect(timeout_sec=10)  
 
    print('Discovering services...')
    UART.discover(device)
 
    # Once service discovery is complete create an instance of the service
    # and start interacting with it.
    uart = UART(device)
 
    # ------------- BT init done ---------------------------------------
    
    # list of initial MM messages.
    init1 = [16] # 02101300
    init2 = [7,1] # 0307010c00
    motion_neutral = [6,0,0,0] # 02060101010c00
 
    msgList = np.array([init1, init2, motion_neutral])
 
    for msg in msgList:
        msgOut = MM_command(msg)
        print("Sending: ", msgOut)   
 
        #msgOut = str.encode(msgOut) # convert to bytes
        msgOut = binascii.unhexlify(msgOut)
        uart.write(msgOut)
 
        time.sleep(0.5)
 
    # main loop. pick up velocity values at a steady rate separate from 
    # frequency of ROS callbacks and send them over BLE
    while 1:
        # convert velocities to MM values
        # float inputs must be converted to signed 8b ints:
        # inputs are handily float 0...1
 
        limit = 80 # max +/- 127
 
        time.sleep(0.2) # separate thread so hopefully ok?
 
        strafe = int(globs.velX * limit)
        fwd = int(globs.velY * limit)
        turn = int(globs.velTheta * limit)
        print ('strafe: %d fwd: %d turn: %d' % (strafe, fwd, turn))
 
        msgOut = MM_command([6, strafe, fwd, turn]) # 6=motion
        print("Sending: ", msgOut)   
 
        #msgOut = str.encode(msgOut)
        msgOut = binascii.unhexlify(msgOut)
 
        uart.write(msgOut)
 
 
 
 
 
def velocity_received_callback(message):
    """Handle new velocity command message."""
 
    # Extract linear and angular velocities from the message
    globs.velX = message.linear.x
    globs.velY = message.linear.y
    globs.velTheta = message.angular.z
 
 
 
 
# startup:
 
rospy.init_node('mmdriver')
 
# Setup subscriber for velocity twist messages
rospy.Subscriber('cmd_vel', Twist, velocity_received_callback)
 
 
# Get the BLE provider for the current platform.
ble = Adafruit_BluefruitLE.get_provider()
 
# Initialize the BLE system.  MUST be called before other BLE calls!
ble.initialize()
 
# Start the mainloop to process BLE events, and run the provided function in
# a background thread.  When the provided main function stops running, returns
# an integer status code, or throws an error the program will exit.
ble.run_mainloop_with(main)
 

